#ifndef MGARDCONFIG_HPP
#define MGARDCONFIG_HPP

/* MGARD Version Information */
#define MGARD_VERSION_STR   "1.6.0"
#define MGARD_VERSION_MAJOR 1
#define MGARD_VERSION_MINOR 6
#define MGARD_VERSION_PATCH 0
/* #undef MGARD_VERSION_TWEAK */

#define MGARD_FILE_VERSION_STR "1.\
  0.0"
#define MGARD_FILE_VERSION_MAJOR 1
#define MGARD_FILE_VERSION_MINOR 0
#define MGARD_FILE_VERSION_PATCH 0
/* #undef MGARD_FILE_VERSION_TWEAK */

#endif /* MGARDCONFIG_HPP */
