#ifndef INTERNAL_CONST_HH
#define INTERNAL_CONST_HH

#include <cstddef>

// dims_L16
extern const size_t DIM0;
extern const size_t DIM1;
extern const size_t DIM2;
extern const size_t DIM3;
extern const size_t nBLK0;
extern const size_t nBLK1;
extern const size_t nBLK2;
extern const size_t nBLK3;
extern const size_t nDIM;
extern const size_t LEN;
extern const size_t CAP;
extern const size_t RADIUS;

// ebs_L4
extern const size_t EB;
extern const size_t EBr;
extern const size_t EBx2;
extern const size_t EBx2_r;

extern const int B_1d;
extern const int B_2d;
extern const int B_3d;

#endif
