#ifndef _MDR_ENCODER_HPP
#define _MDR_ENCODER_HPP

// #include "GroupedBPEncoder.hpp"
// #include "NegaBinaryBPEncoder.hpp"
// #include "PerBitBPEncoder.hpp"
#include "GroupedBPEncoderGPU.hpp"
#include "GroupedWarpBPEncoderGPU.hpp"
// #include "PerBitBPEncoderGPU.hpp"

#endif
