/*
 * Copyright 2022, Oak Ridge National Laboratory.
 * MGARD-X: MultiGrid Adaptive Reduction of Data Portable across GPUs and CPUs
 * Author: Jieyang Chen (chenj3@ornl.gov)
 * Date: March 17, 2022
 */

#include "mgard-x/CompressionLowLevel/Compressor.hpp"
#include "mgard-x/Hierarchy/Hierarchy.hpp"
#include "mgard-x/RuntimeX/DataStructures/Array.hpp"
#include "mgard-x/RuntimeX/RuntimeX.h"